package dev.fyloz.colorrecipesexplorer.exception

import dev.fyloz.colorrecipesexplorer.model.Material
import dev.fyloz.colorrecipesexplorer.model.MaterialQuantityDto
import org.springframework.http.HttpHeaders
import org.springframework.http.HttpStatus
import org.springframework.http.ResponseEntity
import org.springframework.validation.FieldError
import org.springframework.web.bind.MethodArgumentNotValidException
import org.springframework.web.bind.annotation.ControllerAdvice
import org.springframework.web.bind.annotation.ExceptionHandler
import org.springframework.web.context.request.ServletWebRequest
import org.springframework.web.context.request.WebRequest
import org.springframework.web.servlet.mvc.method.annotation.ResponseEntityExceptionHandler

abstract class RestException(
    val errorCode: String,
    val title: String,
    val status: HttpStatus,
    val details: String,
    val extensions: Map<String, Any> = mapOf()
) : RuntimeException(details) {
    fun buildExceptionBody() = mapOf(
        "type" to errorCode,
        "title" to title,
        "status" to status.value(),
        "detail" to details,

        *extensions.map { it.key to it.value }.toTypedArray()
    )
}

class NotFoundException(
    errorCode: String,
    title: String,
    details: String,
    identifierValue: Any,
    identifierName: String = "id"
) : RestException(
    errorCode = "notfound-$errorCode-$identifierName",
    title = title,
    status = HttpStatus.NOT_FOUND,
    details = details,
    extensions = mapOf(
        identifierName to identifierValue
    )
)

class AlreadyExistsException(
    errorCode: String,
    title: String,
    details: String,
    identifierValue: Any,
    identifierName: String = "id"
) : RestException(
    errorCode = "exists-$errorCode-$identifierName",
    title = title,
    status = HttpStatus.CONFLICT,
    details = details,
    extensions = mapOf(
        identifierName to identifierValue
    )
)

class CannotDeleteException(
    errorCode: String,
    title: String,
    details: String
) : RestException(
    errorCode = "cannotdelete-$errorCode",
    title = title,
    status = HttpStatus.CONFLICT,
    details = details
)

@ControllerAdvice
class RestResponseEntityExceptionHandler : ResponseEntityExceptionHandler() {
    @ExceptionHandler(RestException::class)
    fun handleRestExceptions(exception: RestException, request: WebRequest): ResponseEntity<Any> {
        val finalBody = exception.buildExceptionBody().toMutableMap()
        finalBody["instance"] = (request as ServletWebRequest).request.requestURI

        return handleExceptionInternal(
            exception,
            finalBody,
            HttpHeaders(),
            exception.status,
            request
        )
    }

    override fun handleMethodArgumentNotValid(
        ex: MethodArgumentNotValidException,
        headers: HttpHeaders,
        status: HttpStatus,
        request: WebRequest
    ): ResponseEntity<Any> {
        val errors = hashMapOf<String, String>()
        ex.bindingResult.allErrors.forEach {
            val fieldName = (it as FieldError).field
            val errorMessage = it.defaultMessage!!
            errors[fieldName] = errorMessage
        }
        return ResponseEntity(errors, headers, status)
    }
}
