package dev.fyloz.colorrecipesexplorer.model

import com.fasterxml.jackson.annotation.JsonIgnore
import com.fasterxml.jackson.annotation.JsonProperty
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.validation.NullOrNotBlank
import org.hibernate.annotations.Fetch
import org.hibernate.annotations.FetchMode
import org.springframework.security.core.GrantedAuthority
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder
import org.springframework.security.crypto.password.PasswordEncoder
import java.time.LocalDateTime
import javax.persistence.*
import javax.validation.constraints.NotBlank
import javax.validation.constraints.NotNull
import javax.validation.constraints.Size

private const val EMPLOYEE_ID_NULL_MESSAGE = "Un numéro d'employé est requis"
private const val EMPLOYEE_LAST_NAME_EMPTY_MESSAGE = "Un nom  est requis"
private const val EMPLOYEE_FIRST_NAME_EMPTY_MESSAGE = "Un prénom est requis"
private const val EMPLOYEE_PASSWORD_EMPTY_MESSAGE = "Un mot de passe est requis"
private const val EMPLOYEE_PASSWORD_TOO_SHORT_MESSAGE = "Le mot de passe doit contenir au moins 8 caractères"

@Entity
@Table(name = "employee")
data class Employee(
    @Id
    override val id: Long,

    @Column(name = "first_name")
    val firstName: String = "",

    @Column(name = "last_name")
    val lastName: String = "",

    @JsonIgnore
    val password: String = "",

    @JsonIgnore
    @Column(name = "default_group_user")
    val isDefaultGroupUser: Boolean = false,

    @JsonIgnore
    @Column(name = "system_user")
    val isSystemUser: Boolean = false,

    @ManyToOne
    @JoinColumn(name = "group_id")
    @Fetch(FetchMode.SELECT)
    var group: EmployeeGroup? = null,

    @Enumerated(EnumType.STRING)
    @ElementCollection(fetch = FetchType.EAGER)
    @CollectionTable(name = "employee_permission", joinColumns = [JoinColumn(name = "employee_id")])
    @Column(name = "permission")
    @Fetch(FetchMode.SUBSELECT)
    @get:JsonProperty("explicitPermissions")
    val permissions: MutableSet<EmployeePermission> = mutableSetOf(),

    @Column(name = "last_login_time")
    var lastLoginTime: LocalDateTime? = null
) : Model {
    @get:JsonProperty("permissions")
    val flatPermissions: Set<EmployeePermission>
        get() = permissions
            .flatMap { it.flat() }
            .filter { !it.deprecated }
            .toMutableSet()
            .apply {
                if (group != null) this.addAll(group!!.flatPermissions)
            }

    @get:JsonIgnore
    val authorities: Set<GrantedAuthority>
        get() = flatPermissions.map { it.toAuthority() }.toMutableSet()
}

/** DTO for creating employees. Allows a [password] a [groupId]. */
open class EmployeeSaveDto(
    @field:NotNull(message = EMPLOYEE_ID_NULL_MESSAGE)
    val id: Long,

    @field:NotBlank(message = EMPLOYEE_FIRST_NAME_EMPTY_MESSAGE)
    val firstName: String,

    @field:NotBlank(message = EMPLOYEE_LAST_NAME_EMPTY_MESSAGE)
    val lastName: String,

    @field:NotBlank(message = EMPLOYEE_PASSWORD_EMPTY_MESSAGE)
    @field:Size(min = 8, message = EMPLOYEE_PASSWORD_TOO_SHORT_MESSAGE)
    val password: String,

    val groupId: Long?,

    @Enumerated(EnumType.STRING)
    val permissions: MutableSet<EmployeePermission> = mutableSetOf()
) : EntityDto<Employee>

open class EmployeeUpdateDto(
    @field:NotNull(message = EMPLOYEE_ID_NULL_MESSAGE)
    val id: Long,

    @field:NullOrNotBlank(message = EMPLOYEE_FIRST_NAME_EMPTY_MESSAGE)
    val firstName: String?,

    @field:NullOrNotBlank(message = EMPLOYEE_LAST_NAME_EMPTY_MESSAGE)
    val lastName: String?,

    val groupId: Long?,

    @Enumerated(EnumType.STRING)
    val permissions: Set<EmployeePermission>?
) : EntityDto<Employee>

data class EmployeeLoginRequest(val id: Long, val password: String)

// ==== DSL ====
fun employee(
    passwordEncoder: PasswordEncoder = BCryptPasswordEncoder(),
    id: Long = 0L,
    firstName: String = "firstName",
    lastName: String = "lastName",
    password: String = passwordEncoder.encode("password"),
    isDefaultGroupUser: Boolean = false,
    isSystemUser: Boolean = false,
    group: EmployeeGroup? = null,
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    lastLoginTime: LocalDateTime? = null,
    op: Employee.() -> Unit = {}
) = Employee(
    id,
    firstName,
    lastName,
    password,
    isDefaultGroupUser,
    isSystemUser,
    group,
    permissions,
    lastLoginTime
).apply(op)

fun employeeSaveDto(
    passwordEncoder: PasswordEncoder = BCryptPasswordEncoder(),
    id: Long = 0L,
    firstName: String = "firstName",
    lastName: String = "lastName",
    password: String = passwordEncoder.encode("password"),
    groupId: Long? = null,
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    op: EmployeeSaveDto.() -> Unit = {}
) = EmployeeSaveDto(id, firstName, lastName, password, groupId, permissions).apply(op)

fun employeeUpdateDto(
    id: Long = 0L,
    firstName: String = "firstName",
    lastName: String = "lastName",
    groupId: Long? = null,
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    op: EmployeeUpdateDto.() -> Unit = {}
) = EmployeeUpdateDto(id, firstName, lastName, groupId, permissions).apply(op)

// ==== Exceptions ====
private const val EMPLOYEE_NOT_FOUND_EXCEPTION_TITLE = "Employee not found"
private const val EMPLOYEE_ALREADY_EXISTS_EXCEPTION_TITLE = "Employee already exists"
private const val EMPLOYEE_EXCEPTION_ERROR_CODE = "employee"

fun employeeIdNotFoundException(id: Long) =
    NotFoundException(
        EMPLOYEE_EXCEPTION_ERROR_CODE,
        EMPLOYEE_NOT_FOUND_EXCEPTION_TITLE,
        "An employee with the id $id could not be found",
        id
    )

fun employeeIdAlreadyExistsException(id: Long) =
    AlreadyExistsException(
        EMPLOYEE_EXCEPTION_ERROR_CODE,
        EMPLOYEE_ALREADY_EXISTS_EXCEPTION_TITLE,
        "An employee with the id $id already exists",
        id
    )

fun employeeFullNameAlreadyExistsException(firstName: String, lastName: String) =
    AlreadyExistsException(
        EMPLOYEE_EXCEPTION_ERROR_CODE,
        EMPLOYEE_ALREADY_EXISTS_EXCEPTION_TITLE,
        "An employee with the name '$firstName $lastName' already exists",
        "$firstName $lastName",
        "fullName"
    )
