package dev.fyloz.colorrecipesexplorer.model

import com.fasterxml.jackson.annotation.JsonProperty
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.exception.RestException
import org.hibernate.annotations.Fetch
import org.hibernate.annotations.FetchMode
import org.springframework.http.HttpStatus
import javax.persistence.*
import javax.validation.constraints.NotBlank
import javax.validation.constraints.NotNull
import javax.validation.constraints.Size

private const val GROUP_ID_NULL_MESSAGE = "Un identifiant est requis"
private const val GROUP_NAME_NULL_MESSAGE = "Un nom est requis"
private const val GROUP_PERMISSIONS_EMPTY_MESSAGE = "Au moins une permission est requise"

@Entity
@Table(name = "employee_group")
data class EmployeeGroup(
    @Id
    @GeneratedValue(strategy = GenerationType.IDENTITY)
    override var id: Long? = null,

    @Column(unique = true)
    override val name: String = "",

    @Enumerated(EnumType.STRING)
    @ElementCollection(fetch = FetchType.EAGER)
    @CollectionTable(name = "group_permission", joinColumns = [JoinColumn(name = "group_id")])
    @Column(name = "permission")
    @Fetch(FetchMode.SUBSELECT)
    @get:JsonProperty("explicitPermissions")
    val permissions: MutableSet<EmployeePermission> = mutableSetOf(),
) : NamedModel {
    @get:JsonProperty("permissions")
    val flatPermissions: Set<EmployeePermission>
        get() = this.permissions
            .flatMap { it.flat() }
            .filter { !it.deprecated }
            .toSet()
}

open class EmployeeGroupSaveDto(
    @field:NotBlank(message = GROUP_NAME_NULL_MESSAGE)
    @field:Size(min = 3)
    val name: String,

    @field:Size(min = 1, message = GROUP_PERMISSIONS_EMPTY_MESSAGE)
    val permissions: MutableSet<EmployeePermission>
) : EntityDto<EmployeeGroup> {
    override fun toEntity(): EmployeeGroup =
        EmployeeGroup(null, name, permissions)
}

open class EmployeeGroupUpdateDto(
    @field:NotNull(message = GROUP_ID_NULL_MESSAGE)
    val id: Long,

    @field:NotBlank(message = GROUP_NAME_NULL_MESSAGE)
    @field:Size(min = 3)
    val name: String,

    @field:Size(min = 1, message = GROUP_PERMISSIONS_EMPTY_MESSAGE)
    val permissions: MutableSet<EmployeePermission>
) : EntityDto<EmployeeGroup> {
    override fun toEntity(): EmployeeGroup =
        EmployeeGroup(id, name, permissions)
}

fun employeeGroup(
    id: Long? = null,
    name: String = "name",
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    op: EmployeeGroup.() -> Unit = {}
) = EmployeeGroup(id, name, permissions).apply(op)

fun employeeGroupSaveDto(
    name: String = "name",
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    op: EmployeeGroupSaveDto.() -> Unit = {}
) = EmployeeGroupSaveDto(name, permissions).apply(op)

fun employeeGroupUpdateDto(
    id: Long = 0L,
    name: String = "name",
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    op: EmployeeGroupUpdateDto.() -> Unit = {}
) = EmployeeGroupUpdateDto(id, name, permissions).apply(op)

// ==== Exceptions ====
private const val EMPLOYEE_NOT_FOUND_EXCEPTION_TITLE = "Employee group not found"
private const val EMPLOYEE_ALREADY_EXISTS_EXCEPTION_TITLE = "Employee group already exists"
private const val EMPLOYEE_EXCEPTION_ERROR_CODE = "employeegroup"

class NoDefaultGroupException : RestException(
    "nodefaultgroup",
    "No default group",
    HttpStatus.NOT_FOUND,
    "No default group cookie is defined in the current request"
)

fun employeeGroupIdNotFoundException(id: Long) =
    NotFoundException(
        EMPLOYEE_EXCEPTION_ERROR_CODE,
        EMPLOYEE_NOT_FOUND_EXCEPTION_TITLE,
        "An employee group with the id $id could not be found",
        id
    )

fun employeeGroupNameNotFoundException(name: String) =
    NotFoundException(
        EMPLOYEE_EXCEPTION_ERROR_CODE,
        EMPLOYEE_NOT_FOUND_EXCEPTION_TITLE,
        "An employee group with the name $name could not be found",
        name,
        "name"
    )

fun employeeGroupNameAlreadyExistsException(name: String) =
    AlreadyExistsException(
        EMPLOYEE_EXCEPTION_ERROR_CODE,
        EMPLOYEE_ALREADY_EXISTS_EXCEPTION_TITLE,
        "An employee group with the name $name already exists",
        name,
        "name"
    )

