package dev.fyloz.colorrecipesexplorer.model

import com.fasterxml.jackson.annotation.JsonIgnore
import dev.fyloz.colorrecipesexplorer.exception.CannotDeleteException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.validation.NullOrNotBlank
import javax.persistence.*
import javax.validation.constraints.Min
import javax.validation.constraints.NotBlank
import javax.validation.constraints.NotNull

private const val MIX_ID_NULL_MESSAGE = "Un identifiant est requis"
private const val MIX_NAME_NULL_MESSAGE = "Un nom est requis"
private const val MIX_RECIPE_NULL_MESSAGE = "Un recette est requise"
private const val MIX_MATERIAL_TYPE_NULL_MESSAGE = "Un type de produit est requis"

private const val MIX_DEDUCT_MIX_ID_NULL_MESSAGE = "Un identifiant de mélange est requis"
private const val MIX_DEDUCT_RATIO_NULL_MESSAGE = "Un ratio est requis"
private const val MIX_DEDUCT_RATION_NEGATIVE_MESSAGE = "Le ratio doit être égal ou supérieur à 0"

@Entity
@Table(name = "mix")
data class Mix(
    @Id
    @GeneratedValue(strategy = GenerationType.IDENTITY)
    override val id: Long?,

    var location: String?,

    @JsonIgnore
    @ManyToOne
    @JoinColumn(name = "recipe_id")
    val recipe: Recipe,

    @ManyToOne
    @JoinColumn(name = "mix_type_id")
    var mixType: MixType,

    @OneToMany(cascade = [CascadeType.ALL], fetch = FetchType.EAGER, orphanRemoval = true)
    @JoinColumn(name = "mix_id")
    var mixMaterials: MutableSet<MixMaterial>,
) : Model

open class MixSaveDto(
    @field:NotBlank(message = MIX_NAME_NULL_MESSAGE)
    val name: String,

    @field:NotNull(message = MIX_RECIPE_NULL_MESSAGE)
    val recipeId: Long,

    @field:NotNull(message = MIX_MATERIAL_TYPE_NULL_MESSAGE)
    val materialTypeId: Long,

    val mixMaterials: Set<MixMaterialDto>?
) : EntityDto<Mix> {
    override fun toEntity(): Mix = throw UnsupportedOperationException()
}

open class MixUpdateDto(
    @field:NotNull(message = MIX_ID_NULL_MESSAGE)
    val id: Long,

    @field:NullOrNotBlank(message = MIX_NAME_NULL_MESSAGE)
    val name: String?,

    val materialTypeId: Long?,

    var mixMaterials: Set<MixMaterialDto>?
) : EntityDto<Mix> {
    override fun toEntity(): Mix = throw UnsupportedOperationException()
}

data class MixDeductDto(
    @field:NotNull(message = MIX_DEDUCT_MIX_ID_NULL_MESSAGE)
    val id: Long,

    @field:NotNull(message = MIX_DEDUCT_RATIO_NULL_MESSAGE)
    @field:Min(value = 0, message = MIX_DEDUCT_RATION_NEGATIVE_MESSAGE)
    val ratio: Float
)

data class MixLocationDto(
    @field:NotNull(message = MIX_DEDUCT_MIX_ID_NULL_MESSAGE)
    val mixId: Long,

    val location: String?
)

// ==== DSL ====
fun mix(
    id: Long? = null,
    location: String? = "location",
    recipe: Recipe = recipe(),
    mixType: MixType = mixType(),
    mixMaterials: MutableSet<MixMaterial> = mutableSetOf(),
    op: Mix.() -> Unit = {}
) = Mix(id, location, recipe, mixType, mixMaterials).apply(op)

fun mixSaveDto(
    name: String = "name",
    recipeId: Long = 0L,
    materialTypeId: Long = 0L,
    mixMaterials: Set<MixMaterialDto>? = setOf(),
    op: MixSaveDto.() -> Unit = {}
) = MixSaveDto(name, recipeId, materialTypeId, mixMaterials).apply(op)

fun mixUpdateDto(
    id: Long = 0L,
    name: String? = "name",
    materialTypeId: Long? = 0L,
    mixMaterials: Set<MixMaterialDto>? = setOf(),
    op: MixUpdateDto.() -> Unit = {}
) = MixUpdateDto(id, name, materialTypeId, mixMaterials).apply(op)

fun mixRatio(
    id: Long = 0L,
    ratio: Float = 1f,
    op: MixDeductDto.() -> Unit = {}
) = MixDeductDto(id, ratio).apply(op)

fun mixLocationDto(
    mixId: Long = 0L,
    location: String? = "location",
    op: MixLocationDto.() -> Unit = {}
) = MixLocationDto(mixId, location).apply(op)

// ==== Exceptions ====
private const val MIX_NOT_FOUND_EXCEPTION_TITLE = "Mix not found"
private const val MIX_CANNOT_DELETE_EXCEPTION_TITLE = "Cannot delete mix"
private const val MIX_EXCEPTION_ERROR_CODE = "mix"

fun mixIdNotFoundException(id: Long) =
    NotFoundException(
        MIX_EXCEPTION_ERROR_CODE,
        MIX_NOT_FOUND_EXCEPTION_TITLE,
        "A mix with the id $id could not be found",
        id
    )

fun cannotDeleteMixException(mix: Mix) =
    CannotDeleteException(
        MIX_EXCEPTION_ERROR_CODE,
        MIX_CANNOT_DELETE_EXCEPTION_TITLE,
        "Cannot delete the mix ${mix.mixType.name} because one or more mixes depends on it"
    )
