package dev.fyloz.colorrecipesexplorer.service

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.colorrecipesexplorer.exception.LowQuantitiesException
import dev.fyloz.colorrecipesexplorer.model.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import kotlin.test.assertEquals
import kotlin.test.assertTrue

class InventoryServiceTest {
    private val materialService: MaterialService = mock()
    private val mixService: MixService = mock()
    private val service = spy(InventoryServiceImpl(materialService, mixService))

    @AfterEach
    fun afterEach() {
        reset(materialService, service)
    }

    // add()

    @Test
    fun `add(materialQuantities) calls add() for each MaterialQuantityDto`() {
        val materialQuantities = listOf(
            materialQuantityDto(materialId = 1, quantity = 1234f),
            materialQuantityDto(materialId = 2, quantity = 2345f),
            materialQuantityDto(materialId = 3, quantity = 3456f),
            materialQuantityDto(materialId = 4, quantity = 4567f)
        )
        val storedQuantity = 2000f

        doAnswer { storedQuantity + (it.arguments[0] as MaterialQuantityDto).quantity }.whenever(service)
            .add(any<MaterialQuantityDto>())

        val found = service.add(materialQuantities)

        materialQuantities.forEach {
            verify(service).add(it)
            assertTrue { found.any { updated -> updated.material == it.material && updated.quantity == storedQuantity + it.quantity } }
        }
    }

    @Test
    fun `add(materialQuantity) updates material's quantity`() {
        withGivenQuantities(0f, 1000f) {
            val updatedQuantity = it + this.quantity
            whenever(materialService.updateQuantity(any(), eq(this.quantity))).doReturn(updatedQuantity)

            val found = service.add(this)

            verify(materialService).updateQuantity(
                argThat { this.id == this@withGivenQuantities.material },
                eq(this.quantity)
            )
            assertEquals(updatedQuantity, found)
        }
    }

    // deductMix()

    @Test
    fun `deductMix() calls deduct() with a collection of MaterialQuantityDto with adjusted quantities`() {
        val material = material(id = 0L, materialType = materialType(usePercentages = false))
        val materialPercents = material(id = 1L, materialType = materialType(usePercentages = true))
        val mixRatio = mixRatio(ratio = 1.5f)
        val mix = mix(
            id = mixRatio.id,
            mixMaterials = mutableSetOf(
                mixMaterial(id = 0L, material = material, quantity = 1000f, position = 0),
                mixMaterial(id = 1L, material = materialPercents, quantity = 50f, position = 1)
            )
        )
        val expectedQuantities = mapOf(
            0L to 1500f,
            1L to 750f
        )

        whenever(mixService.getById(mix.id!!)).doReturn(mix)
        doAnswer {
            (it.arguments[0] as Collection<MaterialQuantityDto>).map { materialQuantity ->
                materialQuantityDto(materialId = materialQuantity.material, quantity = 0f)
            }
        }.whenever(service).deduct(any<Collection<MaterialQuantityDto>>())

        val found = service.deductMix(mixRatio)

        verify(service).deduct(argThat<Collection<MaterialQuantityDto>> {
            this.all { it.quantity == expectedQuantities[it.material] }
        })

        assertEquals(expectedQuantities.size, found.size)
    }

    // deduct()

    @Test
    fun `deduct(materialQuantities) calls deduct() for each MaterialQuantityDto`() {
        val materialQuantities = listOf(
            materialQuantityDto(materialId = 1, quantity = 1234f),
            materialQuantityDto(materialId = 2, quantity = 2345f),
            materialQuantityDto(materialId = 3, quantity = 3456f),
            materialQuantityDto(materialId = 4, quantity = 4567f)
        )
        val storedQuantity = 5000f

        doAnswer { storedQuantity - (it.arguments[0] as MaterialQuantityDto).quantity }.whenever(service)
            .deduct(any<MaterialQuantityDto>())

        val found = service.deduct(materialQuantities)

        materialQuantities.forEach {
            verify(service).deduct(it)
            assertTrue { found.any { updated -> updated.material == it.material && updated.quantity == storedQuantity - it.quantity } }
        }
    }

    @Test
    fun `deduct(materialQuantities) throws LowQuantitiesException when there is not enough inventory of a given material`() {
        val materialQuantities = listOf(
            materialQuantityDto(materialId = 1, quantity = 1234f),
            materialQuantityDto(materialId = 2, quantity = 2345f),
            materialQuantityDto(materialId = 3, quantity = 3456f),
            materialQuantityDto(materialId = 4, quantity = 4567f)
        )
        val inventoryQuantity = 3000f
        val lowQuantities = materialQuantities.filter { it.quantity > inventoryQuantity }

        materialQuantities.forEach {
            withGivenQuantities(inventoryQuantity, it)
        }

        val exception = assertThrows<LowQuantitiesException> { service.deduct(materialQuantities) }
        assertTrue { exception.materialQuantities.containsAll(lowQuantities) }
    }

    @Test
    fun `deduct(materialQuantity) updates material's quantity`() {
        withGivenQuantities(5000f, 1000f) {
            val updatedQuantity = it - this.quantity
            whenever(materialService.updateQuantity(any(), eq(-this.quantity))).doReturn(updatedQuantity)

            val found = service.deduct(this)

            verify(materialService).updateQuantity(
                argThat { this.id == this@withGivenQuantities.material },
                eq(-this.quantity)
            )
            assertEquals(updatedQuantity, found)
        }
    }

    @Test
    fun `deduct(materialQuantity) throws LowQuantityException when there is not enough inventory of the given material`() {
        withGivenQuantities(0f, 1000f) {
            val exception = assertThrows<NotEnoughInventoryException> { service.deduct(this) }
            assertEquals(this, exception.materialQuantity)
        }
    }

    private fun withGivenQuantities(
        stored: Float,
        quantity: Float,
        materialId: Long = 0L,
        test: MaterialQuantityDto.(Float) -> Unit = {}
    ) {
        val materialQuantity = materialQuantityDto(materialId = materialId, quantity = quantity)

        withGivenQuantities(stored, materialQuantity, test)
    }

    private fun withGivenQuantities(
        stored: Float,
        materialQuantity: MaterialQuantityDto,
        test: MaterialQuantityDto.(Float) -> Unit = {}
    ) {
        val material = material(id = materialQuantity.material, inventoryQuantity = stored)

        whenever(materialService.getById(material.id!!)).doReturn(material)

        materialQuantity.test(stored)
    }
}
